// ==========================================================================
// $Source: /var/lib/cvs/Givaro/src/kernel/zpz/givzpz.h,v $
// Copyright(c)'1994-2009 by The Givaro group
// This file is part of Givaro.
// Givaro is governed by the CeCILL-B license under French law
// and abiding by the rules of distribution of free software.
// see the COPYRIGHT file for more details.
// Authors: T. Gautier
// $Id: givzpz.h,v 1.8 2011-02-02 16:23:56 briceboyer Exp $
// ==========================================================================

/*!@file ring/modular.h
 * @brief   Family of arithmetics over Zpz (\f$\mathbf{Z}/p\mathbf{Z}\f$).
 */

#ifndef __GIVARO_modular_H
#define __GIVARO_modular_H

#include <givaro/givconfig.h>

#include "givaro/modular-integral.h"
#include "givaro/modular-floating.h"
#include "givaro/modular-integer.h"
#include "givaro/modular-inttype.h"
#include "givaro/modular-log16.h"
#include "givaro/modular-ruint.h"

#endif

/* -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:sts=4:sw=4:ts=4:et:sr:cino=>s,f0,{0,g0,(0,\:0,t0,+0,=s
