// ==========================================================================
// Copyright(c)'1994-2015 by The Givaro group
// This file is part of Givaro.
// Givaro is governed by the CeCILL-B license under French law
// and abiding by the rules of distribution of free software.
// see the COPYRIGHT file for more details.
// Authors: Pascal Giorgi <pascal.giorgi@ens-lyon.fr>
//          Clement Pernet <Clement.Pernet@imag.fr>
//          Brice Boyer (briceboyer) <boyer.brice@gmail.com>
//          Alexis Breust (adpated from FFLAS-FFPACK)
// ==========================================================================

/*! @file field/modular-balanced.h
 * @ingroup field
 * @brief Balanced representation of <code>Z/mZ</code> over several types .
 */

#ifndef __GIVARO_modular_balanced_H
#define __GIVARO_modular_balanced_H

#include <givaro/givconfig.h>

#include "givaro/modular-balanced-double.h"
#include "givaro/modular-balanced-float.h"
#include "givaro/modular-balanced-int32.h"
#include "givaro/modular-balanced-int64.h"

#endif // __GIVARO_modular_balanced_double_H

/* -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:sts=4:sw=4:ts=4:et:sr:cino=>s,f0,{0,g0,(0,\:0,t0,+0,=s
